;;;; codewalk.scm
;
; Copyright (c) 2006, Felix L. Winkelmann
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
; conditions are met:
;
;   Redistributions of source code must retain the above copyright notice, this list of conditions and the following
;     disclaimer. 
;   Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
;     disclaimer in the documentation and/or other materials provided with the distribution. 
;   Neither the name of the author nor the names of its contributors may be used to endorse or promote
;     products derived from this software without specific prior written permission. 
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
; AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
; CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
; SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
; POSSIBILITY OF SUCH DAMAGE.
;
; Send bugs, suggestions and ideas to: 
;
; felix@call-with-current-continuation.org
;
; Felix L. Winkelmann
; Unter den Gleichen 1
; 37130 Gleichen
; Germany


#+(not csi)
(declare
  (export expansion-context expansion-context? macroexpand ##sys#macroexpand-0
	  expand expand/context macroexpand ##sys#macroexpand-1-local install-toplevel-expander
	  ##sys#compiler-toplevel-macroexpand-hook ##sys#interpreter-toplevel-macroexpand-hook) )


(define-record expansion-context
  evaluate?
  compilation?
  unsafe?
  raw?
  classes
  prehook
  env)


(define (expansion-context #!key evaluate (compilation (memq #:compiling ##sys#features)) unsafe (classes #t) (cooked evaluate)
			   prehook (environment '()))
  (make-expansion-context evaluate compilation unsafe (not cooked) classes prehook environment) )

(define *mark* (list 'mark))

(define (expand expr #!optional hook . args)
  (expand0 expr hook (expansion-context) args) )

(define (expand/context expr context #!optional hook . args)
  (expand0 expr hook context args) )

(define (expand0 expr hook context args)
  (let ((eval? (expansion-context-evaluate? context))
	(unsafe? (expansion-context-unsafe? context))
	(classes (expansion-context-classes context))
	(raw? (expansion-context-raw? context))
	(pre (expansion-context-prehook context))
	(env (expansion-context-env context))
	(comp? (expansion-context-compilation? context)) )
    (define (lookup var e)
      (ormap (cut memq var <>) e) )
    (define (defined? var e) (lookup var e))
    (define (macroexpand-1-checked x e)
      (let ([x2 (##sys#macroexpand-1-local x '())])
	(if (pair? x2)
	    (let ([h (##sys#slot x2 0)])
	      (if (and (eq? h 'let) (not (defined? 'let e)))
		  (let ([next (cdr x2)])
		    (if (and (pair? next) (symbol? (car next)))
			(macroexpand-1-checked x2 e)
			x2) )
		  x2) )
	    x2) ) )
    (define (mark x) (list *mark* x))
    (define (dowalk c x xo e def d args)
      (call/cc
       (lambda (return)
	 (cond (d (def xo e args))
	       ((and hook (or (eq? #t classes) (memq c classes)))
		(apply 
		 hook x c
		 (lambda (x3 . args2) 
		   (if (eq? x3 x)
		       (return (def xo e args))
		       (walk x3 e d args2) ) )
		 e
		 mark
		 args) )
	       (else (def x e args) ) ) ) ) )
    (define (same x e args) x)
    (define (same-id x e args) `(,(car x) ,(walk (cadr x) e #t args)))
    (define (walk x e d args)
      ;;(pp `(WALK: ,d ,x ,args))
      (define (walk0 x args)
	(cond ((symbol? x) (dowalk 'ref x x e same d args))
	      ((or (number? x) (boolean? x) (char? x) (eof-object? x) (string? x))
	       (dowalk 'literal x x e same d args) )
	      ((and (pair? x) (symbol? (car x)))
	       (let ((head (car x))
		     (body (cdr x)) )
		 (if (defined? head e)
		     (walk-call x e d args)
		     (let ((x2 (macroexpand-1-checked x e)))
		       (if (eq? x2 x)
			   (case head
			     ((quote)
			      (##sys#check-syntax 'quote x '(quote _) #f)
			      (dowalk 'quoted-literal x x e same d args) )
			     ((##core#global-ref)
			      (dowalk 'ref (cadr x) x e same d args) )
			     ((##core#check)
			      (if raw?
				  (dowalk 'special x x e same-id d args) 
				  (walk (if unsafe? '#t (cadr x)) e d args) ) )
			     ((##core#immutable)
			      (if raw?
				  (dowalk 'special x x e same-id d args) 
				  (walk (cadr x) e d args) ) )
			     ((##core#undefined) 
			      (dowalk 'undefined x x e same d args) )
			     ((if)
			      (##sys#check-syntax 'if x '(if _ _ . #(_)) #f)
			      (dowalk 
			       'if x x e
			       (lambda (x e args)
				 `(if ,(walk (cadr x) e d args)
				      ,(walk (caddr x) e d args)
				      ,@(if (pair? (cdddr x))
					    (list (walk (cadddr x) e d args))
					    '() ) ) )
			       d args) )
			     ((begin)
			      (##sys#check-syntax 'begin x '(begin . #(_ 0)) #f)
			      (dowalk 
			       'begin x x e 
			       (lambda (x e args)
				 `(begin ,@(map (cut walk <> e d args) (cdr x))) )
			       d args) )
			     ((set! ##core#set!)
			      (##sys#check-syntax 'set! x '(_ _ _) #f)
			      (dowalk
			       'set!
			       (if (eq? '##core#set! (car x)) `(set! ,@(cdr x)) x)
			       x e
			       (lambda (x e args)
				 `(set! ,(cadr x) ,(walk (caddr x) e d args)) )
			       d args) )
			     ((let)
			      (##sys#check-syntax 'let x '(let #((variable _) 0) . #(_ 1)) #f)
			      (let* ((bindings (cadr x))
				     (vars (map car bindings)) 
				     (e2 (cons vars e))
				     (body (##sys#canonicalize-body (cddr x) (cut defined? <> e2) '())) 
				     (x2 `(let ,bindings ,body)) )
				(dowalk
				 'let x2 x2 e2
				 (lambda (x e args)
				   `(let ,(map (lambda (b) (list (car b) (walk (cadr b) e d args))) bindings)
				      ,(walk body e d args) ) )
				 d args) ) )
			     ((lambda)
			      (##sys#check-syntax 'lambda x '(lambda lambda-list . #(_ 1)) #f)
			      (let ((llist (cadr x))
				    (body (cddr x)) )
				(when (##sys#extended-lambda-list? llist)
				  (set!-values 
				   (llist body) 
				   (##sys#expand-extended-lambda-list 
				    llist body
				    ##sys#syntax-error-hook) ) )
				(##sys#decompose-lambda-list
				 llist
				 (lambda (vars argc rest)
				   (let* ((e2 (cons vars e))
					  (body (##sys#canonicalize-body body (cut defined? <> e2) '())) 
					  (x2 `(lambda ,llist ,body)) )
				     (dowalk
				      'lambda x2 x2 e2
				      (lambda (x e args)
					`(lambda ,llist ,(walk (caddr x) e d args)) )
				      d args) ) ) ) ) )
			     ((##core#loop-lambda)
			      (walk `(lambda ,@(cdr x)) e d args) )
			     ((##core#named-lambda) 
			      (walk `(lambda ,@(cddr x)) e d args) )
			     ((##core#require-for-syntax)
			      (dowalk
			       'require x x e
			       (if raw?
				   same
				   (lambda (x e args)
				     (let ((ids (if eval?
						    (map eval (cdr x)) 
						    (map (lambda (id)
							   (if (and (pair? id) (eq? 'quote (car id)))
							       (cadr id)
							       (error "unable to figure out requirement id" id) ) )
							 (cdr x) ) ) ) )
				       (when eval? (apply ##sys#require ids))
				       (let ((rs (##sys#lookup-runtime-requirements ids)))
					 (walk
					  (if (null? rs)
					      '(##core#undefined)
					      `(##sys#require ,@(map (lambda (x) `',x) rs)) )
					  e d args) ) ) ) )
			       d args) )
			     ((##core#require-extension)
			      (dowalk
			       'require x x e
			       (if raw? 
				   same
				   (lambda (x e args)
				     (walk
				      (let loop ((ids (cdr x)))
					(if (null? ids)
					    '(##core#undefined)
					    (let-values (((exp _) (##sys#do-the-right-thing (cadar ids) comp?)))
					      `(begin ,exp ,(loop (cdr ids))) ) ) ) ; XXX really insert runtime-requirements?
				      e d args) ) )
			       d args) )
			     ((##core#elaborationtimeonly ##core#compiletimeonly)
			      (dowalk 
			       'special x x e
			       (if raw?
				   same-id
				   (lambda (x e args)
				     (when eval? (eval (cadr x)))
				     (walk '(##core#undefined) e d args) ) )
			       d args) )
			     ((##core#elaborationtimetoo ##core#compiletimetoo)
			      (dowalk 
			       'special x x e
			       (if raw?
				   same-id
				   (lambda (x e args)
				     (when eval? (eval (cadr x)))
				     (walk (if comp? (cadr x) '(##core#undefined)) e d args) ) )
			       d args) )
			     ((##core#declare ##core#callunit ##core#define-inline ##core#define-constant ##core#inline_ref
					      ##core#inline_loc_ref ##core#foreign-callback-wrapper
					      ##core#define-foreign-variable ##core#define-foreign-type
					      ##core#foreign-lambda ##core#foreign-callback-lambda
					      ##core#foreign-primitive ##core#foreign-lambda* ##core#foreign-callback-lambda*
					      ##core#define-external-variable)
			      (dowalk 'compiler-form x x e same d args) )
			     ((##core#inline ##core#inline_allocate ##core#inline_update ##core#inline_loc_update)
			      (dowalk 
			       'compiler-form x x e
			       (lambda (x e args)
				 `(##core#inline ,(cadr x) ,@(map (cut walk <> e d args) (cddr x))) )
			       d args) )
			     ((##core#app)
			      (walk-call (cdr x) e d args) )
			     (else (walk-call x e d args)) )
			   (walk x2 e d args) ) ) ) ) )
	      ((null? x) (syntax-error 'expand "invalid syntax" x))
	      (else (walk-call x e d args)) ) )
      (if pre
	  (apply pre x e (lambda (x . args) (walk0 x args)) args)
	  (walk0 x args) ) )
    (define (walk-call x e d args)
      (if (eq? *mark* (car x))
	  (walk (cadr x) e #t args) 
	  (dowalk 
	   'app x x e
	   (lambda (x e args)
	     (map (cut walk <> e d args) x) )
	   d args) ) )
    (walk 
     (##sys#compiler-toplevel-macroexpand-hook expr)
     env #f args) ) )

(define (install-toplevel-expander expander #!optional override-macroexpand)
  (define (expander2 x)
    ;; bind temporarily to avoid endless recursion:
    (fluid-let ((##sys#compiler-toplevel-macroexpand-hook identity))
      (expander x) ) )
  (set! ##sys#compiler-toplevel-macroexpand-hook expander2)
  (set! ##sys#interpreter-toplevel-macroexpand-hook expander2)
  (when override-macroexpand
    (set! macroexpand (lambda (exp . me) (expander exp)))
    (when (##sys#symbol-has-toplevel-binding? '##sys#macroexpand-0) ; for backwards compatibility
      (set! ##sys#macroexpand-0 (lambda (x me) x)) ) 
    #;(set! ##sys#macroexpand-1-local (lambda (x me) x)) ) )
